"""Top-level package for Kingdon."""

__author__ = """Martin Roelfs"""
__email__ = 'martinroelfs@yahoo.com'
__version__ = '1.0.3'

from sympy import symbols

from kingdon.algebra import Algebra
from kingdon.multivector import MultiVector
from kingdon.matrixreps import expr_as_matrix
